import uvicorn
from argparse import ArgumentParser, Namespace
from croydon.command import Command


class Prod(Command):

    NAME = "prod"
    HELP = "run production server"
    ASYNC_RUN = False

    host: str
    port: int

    async def setup(self, args: Namespace) -> None:
        self.host = args.host
        self.port = args.port

    def run_sync(self):
        uvicorn.run("app:app", host=self.host, port=self.port, reload=False)

    def init_argument_parser(self, parser: ArgumentParser) -> None:
        parser.add_argument("-P", "--port", type=int, help="port to listen on", default=8000)
        parser.add_argument("-H", "--host", type=str, help="host to bind to", default="127.0.0.1")



import asyncio
from abc import ABC
from typing import Optional
from unittest import IsolatedAsyncioTestCase
from pydantic import BaseModel as PydanticModel
from ..models.base_model import BaseModel
from ..models.fields import StringField, ListField
from ..models.index import Index, IndexDirection, IndexKey
from ..decorators import api_field


class TestBaseModel(IsolatedAsyncioTestCase):

    def test_empty(self):
        model = BaseModel()
        self.assertIsNone(model.id)
        self.assertCountEqual(model._fields, ["id"])

    def test_not_empty(self):
        class Model(BaseModel):
            field = StringField()

        model = Model({"field": "value"})
        self.assertEqual(model.field, "value")
        self.assertCountEqual(model._fields, ["id", "field"])

    def test_collection_name(self):
        class Model(BaseModel):
            COLLECTION = "models"

        bm = BaseModel()
        self.assertEqual(bm.collection, "base_model")

        m = Model()
        self.assertEqual(m.collection, "models")

    def test_rejected_fields(self):
        class Model(BaseModel):
            string_f = StringField()
            string_rf = StringField(rejected=True)

        model = Model()
        self.assertCountEqual(model._fields, ["id", "string_f", "string_rf"])
        self.assertTrue(model._fields["string_rf"].rejected)

    def test_restricted_fields(self):
        class Model(BaseModel):
            string_f = StringField()
            string_rf = StringField(restricted=True)

        model = Model()
        self.assertCountEqual(model._fields, ["id", "string_f", "string_rf"])
        self.assertTrue(model._fields["string_rf"].restricted)

    def test_indexes(self):
        class Model(BaseModel):
            string_f = StringField()
            string_if = StringField(index=True)
            string_uf = StringField(unique=True)
            string_duf = StringField(index=IndexDirection.DESCENDING, unique=True)

        model = Model()

        self.assertCountEqual(
            model._indexes,
            [
                Index(keys=[IndexKey(key="string_if", spec=IndexDirection.ASCENDING)]),
                Index(
                    keys=[IndexKey(key="string_uf", spec=IndexDirection.ASCENDING)],
                    options={"unique": True}
                ),
                Index(
                    keys=[IndexKey(key="string_duf", spec=IndexDirection.DESCENDING)],
                    options={"unique": True}
                ),
            ],
        )

    def test_to_dict(self):
        class Model(BaseModel):
            field1 = StringField()
            field2 = StringField()
            field3 = StringField(restricted=True)

        model = Model({"field1": "value1", "field2": "value2", "field3": "value3"})

        self.assertDictEqual(
            model.to_dict(), {"id": None, "field1": "value1", "field2": "value2"}
        )

        self.assertDictEqual(
            model.to_dict(fields=["field1", "field2", "field3"]),
            {"field1": "value1", "field2": "value2"},
        )

        self.assertDictEqual(
            model.to_dict(include_restricted=True),
            {"id": None, "field1": "value1", "field2": "value2", "field3": "value3"},
        )

        self.assertDictEqual(
            model.to_dict(fields=["field1", "field3"], include_restricted=True),
            {"field1": "value1", "field3": "value3"},
        )

        self.assertDictEqual(
            model.to_dict(
                fields=["field1", "field3", "bizzare"], include_restricted=True
            ),
            {"field1": "value1", "field3": "value3"},
        )

    async def test_api_field_methods(self):
        class Model(BaseModel):
            a = StringField()

            async def b(self):
                await asyncio.sleep(0.1)
                return "b"

            @api_field
            async def c(self):
                await asyncio.sleep(0.1)
                return "c"

        m = Model.create(a="a")
        self.assertDictEqual(m.to_dict(), {"id": None, "a": "a"})
        self.assertDictEqual(await m.to_dict_ext(), {"id": None, "a": "a"})
        self.assertDictEqual(
            await m.to_dict_ext(fields=["id", "a", "b", "c"]),
            {"id": None, "a": "a", "c": "c"},
        )

    async def test_from_pydantic(self):
        class Model(BaseModel):
            a = StringField(default="default_a")
            b = StringField(default="default_b")

        class PydanticPayload(PydanticModel):
            a: Optional[str]
            b: Optional[str]

        p = PydanticPayload(a="a")
        m = Model.from_pydantic(p)

        self.assertEqual(m.a, "a")
        self.assertEqual(m.b, "default_b")  # unset value in input must not override defaults

    async def test_initial_state_immutability(self):
        class Model(BaseModel):
            a = ListField(default=list)

        m = Model.create(a=[1, 2, 3])
        await m.save()

        self.assertCountEqual([1, 2, 3], m.a)
        self.assertCountEqual([1, 2, 3], m._initial_state["a"])

        m.a.append(5)
        self.assertCountEqual([1, 2, 3, 5], m.a)
        self.assertCountEqual([1, 2, 3], m._initial_state["a"])

        await m.save()
        self.assertCountEqual([1, 2, 3, 5], m._initial_state["a"])

    async def test_computed_fields(self):
        """
        This test was made due to an unpleasant bug when all the models
        ended up having all the computed fields from the entire project due
        to inaccurate modifying of the _computed_fields property
        :return:
        """
        class Person(BaseModel, ABC):
            first_name = StringField(required=True)
            last_name = StringField(required=True)

            @api_field
            def full_name(self) -> str:
                return f"{self.first_name} {self.last_name}"

        class Singer(Person):
            genre = StringField(default="blues")

        class AnotherModel(BaseModel):

            @api_field
            def custom(self) -> str:
                return "custom"

        self.assertCountEqual([], BaseModel._computed_fields.keys())
        self.assertCountEqual(["full_name"], Person._computed_fields.keys())
        self.assertCountEqual(["full_name"], Singer._computed_fields.keys())
        self.assertCountEqual(["custom"], AnotherModel._computed_fields.keys())

    async def test_exposed_fields(self):

        class Person(BaseModel, ABC):
            first_name = StringField(required=True)
            last_name = StringField(required=True)
            password = StringField(required=True, restricted=True)

            @api_field
            def full_name(self) -> str:
                return f"{self.first_name} {self.last_name}"

        self.assertCountEqual(Person.exposed_base_fields(), ["id", "first_name", "last_name"])
        self.assertCountEqual(Person.exposed_fields(), ["id", "first_name", "last_name", "full_name"])



from enum import Enum
from dataclasses import dataclass
from typing import Type, TYPE_CHECKING

if TYPE_CHECKING:
    from .base_model import TBaseModel


class OnDestroy(Enum):
    CASCADE = "cascade"
    RAISE = "raise"
    DETACH = "detach"


@dataclass
class ModelReference:
    ref_class: Type["TBaseModel"]
    ref_field: str
    on_destroy: OnDestroy

    def __hash__(self):
        return hash(f"{self.ref_class.__name__}.{self.ref_field}")



import itertools
from bson import ObjectId
from croydon.models.submodel import StorableSubmodel
from croydon.models.fields import Field
from .mongo_mock_test import MongoMockTest
from ..errors import WrongSubmodel, MissingSubmodel, IntegrityError


class TestBaseModel(StorableSubmodel):
    COLLECTION = "test"
    field1 = Field()
    field2 = Field()


class Submodel1(TestBaseModel):
    SUBMODEL = "submodel1"
    field3 = Field()


class Submodel2(TestBaseModel):
    SUBMODEL = "submodel2"
    field4 = Field()


TestBaseModel.register_submodel(Submodel1.SUBMODEL, Submodel1)
TestBaseModel.register_submodel(Submodel2.SUBMODEL, Submodel2)


class TestSubmodel(MongoMockTest):

    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        await TestBaseModel.destroy_all()

    async def test_wrong_input(self):
        with self.assertRaises(WrongSubmodel):
            Submodel1({"_id": ObjectId(), "field1": "value", "submodel": "wrong"})
        with self.assertRaises(MissingSubmodel):
            Submodel1({"_id": ObjectId(), "field1": "value"})
        with self.assertRaises(IntegrityError):
            Submodel1({"field1": "value", "submodel": "my_submodel"})
        with self.assertRaises(WrongSubmodel):
            obj = Submodel1({"field1": "value"})
            obj.submodel = "wrong"
            await obj.save()

    async def test_submodel_field(self):
        obj = Submodel1({})
        self.assertTrue(hasattr(obj, "submodel"))
        self.assertEqual(obj.submodel, Submodel1.SUBMODEL)
        await obj.save()
        await obj.reload()
        self.assertEqual(obj.submodel, Submodel1.SUBMODEL)
        db_obj = await Submodel1.get(obj.id)
        self.assertEqual(db_obj.submodel, Submodel1.SUBMODEL)

    async def test_inheritance(self):
        class Submodel1(TestBaseModel):
            SUBMODEL = "submodel1"

        class Submodel1_1(Submodel1):
            pass

        self.assertEqual(TestBaseModel.collection, Submodel1.collection)
        self.assertEqual(Submodel1.collection, Submodel1_1.collection)
        self.assertEqual(Submodel1.SUBMODEL, Submodel1_1.SUBMODEL)

    async def test_abstract(self):
        with self.assertRaises(IntegrityError):
            TestBaseModel({})

        with self.assertRaises(IntegrityError):
            class C(TestBaseModel):
                pass  # no SUBMODEL
            C({})

        with self.assertRaises(IntegrityError):
            class C(TestBaseModel):
                SUBMODEL = "c"
            Submodel1.register_submodel("c", C)

    async def test_fields_inheritance(self):
        m1 = Submodel1({})
        m2 = Submodel2({})

        self.assertCountEqual(m1._fields, ["id", "submodel", "field1", "field2", "field3"])
        self.assertCountEqual(m2._fields, ["id", "submodel", "field1", "field2", "field4"])

    @staticmethod
    async def _create_objs():
        """Returns two lists of objects. Objects in the same positions only differ in their submodel"""
        values = [1, 2, 3]
        objs1 = [Submodel1({"field1": v, "field2": v}) for v in values]
        objs2 = [Submodel2({"field1": v, "field2": v}) for v in values]
        for obj in itertools.chain(objs1, objs2):
            await obj.save()
        return objs1, objs2

    async def test_isolation_find(self):
        self.maxDiff = None
        objs1, objs2 = await self._create_objs()
        self.assertCountEqual(
            await Submodel1.find().all(),
            objs1,
        )
        self.assertCountEqual(
            await Submodel2.find().all(),
            objs2,
        )
        self.assertCountEqual(
            await TestBaseModel.find().all(),
            objs1 + objs2,
        )

        self.assertCountEqual(
            await Submodel1.find({"field1": objs1[0].field1}).all(),
            [objs1[0]],
        )
        self.assertCountEqual(
            await TestBaseModel.find({"field1": objs1[0].field1}).all(),
            [objs1[0], objs2[0]],
        )

    async def test_isolation_update(self):
        objs1, objs2 = await self._create_objs()
        obj1 = objs2[0]
        obj1.field2 = "new_value"
        await Submodel2.update_many(
            {"field1": obj1.field1},
            {"$set": {"field2": obj1.field2}}
        )
        self.assertCountEqual(
            await TestBaseModel.find({"field2": obj1.field2}).all(),
            [obj1]
        )

        obj1 = objs1[1]
        obj2 = objs2[1]
        obj1.field2 = "newer_value"
        obj2.field2 = "newer_value"
        await TestBaseModel.update_many(
            {"field1": obj1.field1},
            {"$set": {"field2": obj1.field2}}
        )
        self.assertCountEqual(
            await TestBaseModel.find({"field2": obj1.field2}).all(),
            [obj1, obj2]
        )

    async def test_isolation_destroy(self):
        objs1, objs2 = await self._create_objs()
        to_destroy = objs2.pop()
        to_keep = objs1[-1]
        await Submodel2.destroy_many({"field1": to_destroy.field1})
        self.assertListEqual(
            await Submodel2.find({"field1": to_destroy.field1}).all(),
            []
        )
        self.assertListEqual(
            await TestBaseModel.find({"field1": to_destroy.field1}).all(),
            [to_keep]
        )

        to_destroy = objs1[0]
        objs1 = objs1[1:]
        objs2 = objs2[1:]
        await TestBaseModel.destroy_many({"field1": to_destroy.field1})
        self.assertCountEqual(
            await TestBaseModel.find().all(),
            objs1 + objs2,
        )



from setuptools import setup, find_packages


setup(
    name="croydon",
    version="0.14.0",
    description="a micro webframework based on fastapi and motor",
    url="https://gitlab.com/viert/croydon",
    author="Pavel Vorobyov",
    author_email="aquavitale@yandex.ru",
    license="MIT",
    packages=[pkg for pkg in find_packages() if pkg.startswith("croydon")],
    include_package_data=True,
    package_data={"croydon": ["*.txt", "*.py", "*.tmpl"]},
    python_requires=">=3.11",
    install_requires=[
        "fastapi[all]",
        "uvicorn",
        "pymongo",
        "motor",
        "mongomock-motor",
        "aiomcache",
        "pydantic",
        "ipython",
        "requests",
        "bcrypt",
        "deprecation"
    ],
    entry_points={"console_scripts": ["croydon=croydon.command:main"]},
)



from typing import Optional, Any, Dict
from .abc import AbstractCache

STATIC_CACHE: Dict[str, Any] = {}


class SimpleCache(AbstractCache):
    NAME = "SimpleCache"

    async def initialise(self) -> None:
        pass

    async def get(self, key: str) -> Optional[Any]:
        return STATIC_CACHE.get(key)

    async def set(self, key: str, value: Any) -> None:
        STATIC_CACHE[key] = value

    async def has(self, key: str) -> bool:
        return key in STATIC_CACHE

    async def delete(self, key: str) -> bool:
        had = await self.has(key)
        if had:
            del STATIC_CACHE[key]
        return had
